#-----------------------------------------------------------------
# mazeroute.tcl
#-----------------------------------------------------------------
# Defines procedure "mazeroute <netlist>", requiring a .net file as
# an argument.  Attempts a single-pass maze route of the contents
# of the netlist.
#-----------------------------------------------------------------

global Opts

proc mazeroute {netfile} {
   if [catch {open  $netfile r} fnet] {
      set netname [file rootname $netfile]
      if [catch {open  ${netfile}.net r} fnet] {
         puts stderr "Can't read netlist file $netfile"
	 return 1;
      }
   }
   # 1st line of the netlist file is throw-away
   gets $fnet line

   set destnet {}
   while {[gets $fnet line] >= 0} {
      if {$line == ""} {
	 set destnet {}
      } elseif {$destnet == {}} {
	 set destnet $line
      } else {
	 set startnet $line
	 iroute route -dlabel $destnet -slabel $startnet -timeout 3
	 set destnet $startnet
      }
   }
}

#-----------------------------------------------------------------
# Interactive mazerouter GUI.  Shows a list of nets from the
# selected netlist.  Allows one to select the order of routing,
# route individual nets, rip up individual nets, etc.
#
# This GUI more or less replaces the "specialopen netlist" window.
#-----------------------------------------------------------------

proc genmazehelper {} {

   if {![catch {wm deiconify .mazehelper ; raise .mazehelper}]} {return}

   toplevel .mazehelper
   wm protocol .mazehelper WM_DELETE_WINDOW {destroy .mazehelper}

   frame .mazehelper.mazemenu
   frame .mazehelper.unrouted
   frame .mazehelper.transfer
   frame .mazehelper.routed

   pack .mazehelper.mazemenu -side top -anchor w

   button .mazehelper.mazemenu.load -text "Load" -command {loadnetlist}
   label .mazehelper.mazemenu.netlist -text "(no netlist loaded)"

   pack .mazehelper.mazemenu.load -side left
   pack .mazehelper.mazemenu.netlist -side left

   label .mazehelper.unrouted.title -text "Unrouted:"
   listbox .mazehelper.unrouted.contents -background white -selectmode extended

   label .mazehelper.routed.title -text "Routed:"
   listbox .mazehelper.routed.contents -background white -selectmode extended

   button .mazehelper.transfer.ripup -text "<--" -command {ripupnet}
   button .mazehelper.transfer.route -text "-->" -command {routenet}

   pack .mazehelper.transfer.ripup -side top
   pack .mazehelper.transfer.route -side top

   pack .mazehelper.unrouted.title -side top
   pack .mazehelper.unrouted.contents -side top -fill both -expand true

   pack .mazehelper.routed.title -side top
   pack .mazehelper.routed.contents -side top -fill both -expand true

   pack .mazehelper.unrouted -side left -fill both -expand true
   pack .mazehelper.transfer -side left
   pack .mazehelper.routed -side left -fill both -expand true
}

#-----------------------------------------------------------------
# Load a magic-style netlist
#-----------------------------------------------------------------

proc loadnetlist { {netfile {}} } {

   if {$netfile == {}} {
      set netfile [ tk_getOpenFile -filetypes \
	   {{NET {.net {.net}}} {"All files" {*}}}]
   }
   
   if [catch {open  $netfile r} fnet] {
      set netname [file rootname $netfile]
      if [catch {open  ${netfile}.net r} fnet] {
         puts stderr "Can't read netlist file $netfile"
	 return 1;
      }
   }

   # Clear out the listbox contents.
   .mazehelper.unrouted.contents delete 0 end
   .mazehelper.routed.contents delete 0 end


   # 1st line of the netlist file is throw-away
   gets $fnet line

   set currentnet {}
   while {[gets $fnet line] >= 0} {
      if {$line == ""} {
	 if {[llength $currentnet] > 0} {
	    .mazehelper.unrouted.contents insert end $currentnet
            set currentnet {}
	 }
      } else {
	 lappend currentnet $line
      }
   }

   .mazehelper.mazemenu.netlist configure -text $netfile

   close $fnet
}

#-----------------------------------------------------------------
# (Experimental) Place a contact on each network endpoint.  Should
# aid in preventing the router from routing over pins by blocking
# access to the space over every pin that will be routed to.
#-----------------------------------------------------------------

proc contactendpoints { {viatype via} } {
   set viasize [tech drc width $viatype]
   box values 0 0 0 0
   foreach net [.mazehelper.unrouted.contents get 0 end] {
      foreach endpoint $net {
	 goto $endpoint
	 box size ${viasize}i ${viasize}i
	 paint $viatype
      }
   }
}

#-----------------------------------------------------------------
# Load a list of failed routes
#-----------------------------------------------------------------

proc loadfailed { {netfile {}} } {

   if {$netfile == {}} {
      set netfile [ tk_getOpenFile -filetypes \
	   {{FAILED {.failed {.failed}}} {"All files" {*}}}]
   }
   
   if [catch {open  $netfile r} fnet] {
      set netname [file rootname $netfile]
      if [catch {open  ${netfile}.failed r} fnet] {
         puts stderr "Can't read file of failed routes $netfile"
	 return 1;
      }
   }

   # Clear out the listbox contents.
   .mazehelper.unrouted.contents delete 0 end
   .mazehelper.routed.contents delete 0 end


   # Read each line into the "unrouted" list.
   while {[gets $fnet line] >= 0} {
      .mazehelper.unrouted.contents insert end $line
   }

   .mazehelper.mazemenu.netlist configure -text $netfile

   close $fnet
}

#-----------------------------------------------------------------
# Save the list of failed routes
#-----------------------------------------------------------------

proc savefailed { {netfile {}} } {

   set netfile [.mazehelper.mazemenu.netlist cget -text]
   if {$netfile == {}} {
      set netfile [ tk_getOpenFile -filetypes \
	   {{FAILED {.failed {.failed}}} {"All files" {*}}}]
   } else {
      set netname [file rootname $netfile]
      set netname ${netname}.failed
   }
   
   if [catch {open  $netfile w} fnet] {
      set netname [file rootname $netfile]
      if [catch {open  ${netfile}.failed w} fnet] {
         puts stderr "Can't write file of failed routes $netfile"
	 return 1;
      }
   }

   foreach net [.mazehelper.unrouted.contents get 0 end] {
      puts $fnet "$net"
   }

   close $fnet
}

#-----------------------------------------------------------------
# Get the selected network and maze route it
#-----------------------------------------------------------------

proc routenet {} {
   set unroutable {}
   set sellist [.mazehelper.unrouted.contents curselection]
   set startidx [lindex $sellist 0]
   while {[llength $sellist] > 0} {
      set cidx [lindex $sellist 0]
      set rlist [.mazehelper.unrouted.contents get $cidx]
      for {set i 1} {$i < [llength $rlist]} {incr i} {
         set j $i
         incr j -1
         set startnet [lindex $rlist $j]
         set destnet [lindex $rlist $i]
         set rresult [iroute route -slabel $startnet -dlabel $destnet -timeout 3]

	 # break on any failure
	 if {$rresult != "Route success" && \
		$rresult != "Route best before interrupt" && \
		$rresult != "Route already routed"} {
	    break
	 } else {
	    set rresult "success"
	 }
      }

      .mazehelper.unrouted.contents delete $cidx $cidx
      if {$rresult == "success"} {
         .mazehelper.routed.contents insert end $rlist
      } else {
	 # scramble list; we may have better luck routing in a different order
	 set rfirst [lindex $rlist 0]
	 set rlist [lrange $rlist 1 end]
	 lappend rlist $rfirst
	 lappend unroutable $rlist
      }
      set sellist [.mazehelper.unrouted.contents curselection]
   }
   .mazehelper.unrouted.contents selection set $startidx
   foreach badnet $unroutable {
      .mazehelper.unrouted.contents insert $startidx $badnet
   }
}

#-----------------------------------------------------------------
# Get the selected network and remove it
#-----------------------------------------------------------------

proc ripupnet {} {
   set sellist [.mazehelper.routed.contents curselection]
   set startidx [lindex $sellist 0]
   while {[llength $sellist] > 0} {
      set cidx [lindex $sellist 0]
      set rlist [.mazehelper.routed.contents get $cidx]
      set netname [lindex $rlist 0]
      select clear
      set layertype [goto $netname]
      select more box $layertype	;# chunk
      select more box $layertype	;# region
      select more box $layertype	;# net
      delete

      .mazehelper.routed.contents delete $cidx $cidx
      .mazehelper.unrouted.contents insert end $rlist

      set sellist [.mazehelper.routed.contents curselection]
   }
   .mazehelper.routed.contents selection set $startidx
}

#-----------------------------------------------------------------
# Get the selected network and verify the route
#-----------------------------------------------------------------

proc verifynet { {column routed} } {
   set sellist [.mazehelper.${column}.contents curselection]
   set startidx [lindex $sellist 0]
   foreach cidx $sellist {
      set errors 0
      set rlist [.mazehelper.${column}.contents get $cidx]
      set netname [lindex $rlist 0]
      select clear
      set layertype [goto $netname]
      select more box $layertype	;# chunk
      select more box $layertype	;# region
      select more box $layertype	;# net
      set sellist [what -list]
      set sellabels [lindex $sellist 1]
      set labellist {}
      foreach label $sellabels {
	 set labtext [lindex $label 0]
	 set labinst [lindex $label 2]
	 if {$labinst == {}} {
	    set labname ${labtext}
	 } else {
	    set labname ${labinst}/${labtext}
	 }
	 lappend labellist $labname
      }

      # Compare labellist to rlist---they are supposed to be the same!

      foreach entry $labellist {
	 if {[lsearch $rlist $entry] < 0} {
	    puts stderr "ERROR:  Net entry $entry in layout is not in the netlist!"
	    incr errors
	 }
      }
      foreach entry $rlist {
	 if {[lsearch $labellist $entry] < 0} {
	    puts stderr "ERROR:  Net entry $entry in netlist is not in the layout!"
	    incr errors
	 }
      }
      if {$errors == 0} {puts stdout "VERIFIED"}

      # If column is "routed" and we're not verified, move to "unrouted".
      # If column is "unrouted" and we're verified, move to "routed".

      if {"$column" == "routed"} {
         if {$errors > 0} {
            .mazehelper.routed.contents delete $cidx $cidx
            .mazehelper.unrouted.contents insert end $rlist
         }
      } else {
         if {$errors == 0} {
	    .mazehelper.unrouted.contents delete $cidx $cidx
	    .mazehelper.routed.contents insert end $rlist
         }
      }
   }
   .mazehelper.${column}.contents selection set $startidx
}

#-----------------------------------------------------------------
# Reset the maze helper, deleting all routes.
#-----------------------------------------------------------------

proc resetmazehelper {} {
   .mazehelper.routed.contents delete 0 end
   .mazehelper.unrouted.contents delete 0 end
}

#-----------------------------------------------------------------
# Add the "mazehelper" function to the Magic Options
#-----------------------------------------------------------------

proc addmazehelper {optmenu} {
   global Opts
   $optmenu add check -label "Maze Router" -variable Opts(mazeroute) -command \
	{if {$Opts(mazeroute) == 0} {destroy .mazehelper} else {genmazehelper}}
}

#-----------------------------------------------------------------

